#!/bin/sh

set -o pipefail

#----- includes --------------------------------------------------------

# include platform file
source /etc/platform  2>/dev/null || true

# platform default values
EEPROM="/sys/class/i2c-dev/i2c-0/device/0-0050/eeprom"
[ "${platform_type}" = '' -o "${platform_type}" = '%type%' ] && platform_type='CS141BL'
[ "${platform_oem}"                    = '' ] && platform_oem='012'
[ "${platform_serial}"                 = '' ] && platform_serial="$(dd if="${EEPROM}" skip=0 bs=1 count=11 2>/dev/null)"
[ "${platform_macaddr}"                = '' ] && platform_macaddr="$(dd if="${EEPROM}" skip=256 bs=1 count=18 2>/dev/null | sed -e 's/://g' | egrep -o '[0-9a-fA-F]+')"
[ "${platform_feature_typ}"            = '' ] && platform_feature_typ='bch16'
[ "${platform_feature_dtb}"            = '' ] && platform_feature_dtb='budget'
[ "${platform_feature_ext_dip_switch}" = '' ] && platform_feature_ext_dip_switch='no'
unset EEPROM

platform_macaddr="$(echo -n "${platform_macaddr}" | sed -e 's/://g' | egrep -o '[0-9a-fA-F]+')"
MAC="${platform_macaddr}"
MAC="$(printf "%s:%s:%s:%s:%s:%s\n" "${MAC:0:2}" "${MAC:2:2}" "${MAC:4:2}" "${MAC:6:2}" "${MAC:8:2}" "${MAC:10:2}")"

#----- data ------------------------------------------------------------

if [ "$(basename "${0}")" = 'flashnand.sh' ]; then
   source /etc/config
   MODE='flash'
   BASE='/root/install/'
   ARCH=''
   FTP="ftp://${IR_IPSRV}/${IR_LOAD_PATH}/"
elif [ "$(basename "${0}")" = 'install.sh' ]; then
   MODE='update'
   BASE=''
   ARCH="$(ls /var/www/hiawatha/upload/update[0-9][0-9][0-9].tar.gz 2>/dev/null | head -1)"
   FTP=''
else
   MODE=''
   BASE=''
   ARCH=''
   FTP=''
fi

#----- const -----------------------------------------------------------

MTD_SPL='/dev/mtd0 /dev/mtd1 /dev/mtd2 /dev/mtd3'
MTD_UBOOT='/dev/mtd4'
MTD_U_CNF='/dev/mtd5'
MTD_1_KNL='/dev/mtd6'
MTD_1_DTB='/dev/mtd8'
MTD_2_KNL='/dev/mtd7'
MTD_2_DTB='/dev/mtd9'
MTD_OOPS='/dev/mtd10'
MTD_UBI='/dev/mtd11'

FILE_MNT='/mnt'

FILE_SPL="${BASE}generex_nand_${platform_feature_typ}_uart4-MLO"
FILE_UBOOT="${BASE}generex_nand_${platform_feature_typ}_uart4-u-boot.img"
FILE_ENV="${BASE}env.bin"
FILE_KNL="${BASE}zImage"
FILE_DTB="${BASE}am335x-generex_${platform_feature_typ}-${platform_feature_dtb}.dtb"
FILE_ROOTFS="${BASE}rootfs.tar.gz"
FILE_OEM="${BASE}oem-${platform_oem}.tar.gz"

FILE_PLAT='/etc/platform'

ENV_MOD_VER="4"
ENV_MOD_DUAL="
#----- nand boot logic -------------------------------------------------
envver     ${ENV_MOD_VER}
ethaddr    ${MAC}
ipaddr     10.10.10.10
ncip       10.10.10.11
bootm_size 0x08000000
fdtaddr    0x81000000
loadaddr   0x82000000
bootcmd    if gpio input 58 || gpio input 59; then run nand_pre; else if ping \${ncip}; then setenv stdout nc; setenv stdin nc; echo \"\\\\cU-Boot# \"; fi; if sleep 1; then run nand_pre; fi; fi
nand_pre   run nand_pre1
nand_pre1  if gpio input 5; then run nand_boot2; fi; if gpio input 58 && gpio input 59; then run nand_boot2; fi; run nand_boot1
nand_pre2  if gpio input 5; then run nand_boot1; fi; if gpio input 58 && gpio input 59; then run nand_boot1; fi; run nand_boot2
nand_boot
nand_boot1 nand read \${loadaddr} 0x00280000 0x500000; nand read \${fdtaddr} 0x00c80000 0x20000; run nandargs1; bootz \${loadaddr} - \${fdtaddr}
nand_boot2 nand read \${loadaddr} 0x00780000 0x500000; nand read \${fdtaddr} 0x00ca0000 0x20000; run nandargs2; bootz \${loadaddr} - \${fdtaddr}
console    ttyO4,115200n8
nandargs
nandargs1  setenv bootargs console=\${console} consoleblank=0 quiet ubi.mtd=11 root=ubi0:rootfs rw rootfstype=ubifs rootwait
nandargs2  setenv bootargs console=\${console} consoleblank=0 quiet ubi.mtd=11 root=ubi0:rootfs2 rw rootfstype=ubifs rootwait
"

#----- functions -------------------------------------------------------

leds_work () {
   [ -x /usr/bin/led ] && led red   blink_slow
   [ -x /usr/bin/led ] && led green off
}

leds_done () {
   [ -x /usr/bin/led ] && led red   on
   if [ "${MODE}" = 'flash' ]; then
      [ -x /usr/bin/led ] && led green on
      [ -x /usr/bin/led ] && led red  off
   else
      [ -x /usr/bin/led ] && led green off
   fi
}

leds_error () {
   [ -x /usr/bin/led ] && led red   blink_fast
   [ -x /usr/bin/led ] && led green off
}

pout () {
   echo -e "$@"
   [ "${MODE}" = 'update' ] && sync
}

pok () {
   echo -e ":: 0" "$@"
   sync
   leds_done
   true
}

perr () {
   echo -e "error:" "$@" 1>&2
   if [ "${MODE}" = 'update' ]; then
      echo ":: 1 update failed." 1>&2
   else
      echo ":: 1 first installation failed." 1>&2
   fi
   sync
   leds_error
   # do not use internal funtions here because of possible recursion!
   mountpoint -q "${FILE_MNT}" && umount "${FILE_MNT}"
   [ "${MODE}" = 'flash' -a -c "/dev/ubi0" ] && ubidetach --dev-path="${MTD_UBI}"
   exit 1
}

dev_erase () {
   pout "erase block device ${1} ..."
   flash_erase --quiet "${1}" 0 0 || perr "cannot erase block device ${1}"
}

dev_write () {
   pout "erase and write block device ${1} with ${2} ..."
   [ -f "${2}" ] || perr "file ${2} does not exist"
   flash_erase --quiet "${1}" 0 0 || perr "cannot erase block device ${1}"
   nandwrite --markbad --pad --quiet "${1}" "${2}" || perr "cannot write block device ${1} with ${2}"
}

#only used during mode=update (not during install(mode=flash))!
dev_update_write () {
   pout "erase and write block device ${1} with ${ARCH}:${2} ..."
   flash_erase --quiet "${1}" 0 0 || perr "cannot erase block device ${1}"
   ( gunzip -c "${ARCH}" | tar -xOf - "${2}" | nandwrite --markbad --pad --quiet "${1}" ) ||
      perr "cannot write block device ${1} with ${ARCH}:${2}"
}

ubi_format () {
   pout "format block device ${MTD_UBI} as ubi device ..."
   ubiformat "${MTD_UBI}" --yes --quiet || perr "cannot format block device ${MTD_UBI} as ubi device"
}

ubi_attach () {
   pout "attach block device ${MTD_UBI} as ubi device /dev/ubi0 ..."
   ubiattach --devn=0 --dev-path="${MTD_UBI}" >/dev/null || perr "cannot attach block device ${MTD_UBI} as ubi device /dev/ubi0"
}

ubi_exists () {
   ubinfo --devn=0 --name "$1" &>/dev/null
}

ubi_vol_id () {
   local id="$(ubinfo --devn=0 --name="${1}" 2>/dev/null | egrep '^Volume ID:' | egrep -o '[0-9]+' | head -n 1)"
   [ -z "${id}" ] && perr "cannot get volume id for '${1}'"
   echo "${id}"
}

ubi_mkvol () {
   if [ "${1}" = 'auto' ]; then
      pout "create ubifs volume ${2}(${3}) ..."
      ubimkvol /dev/ubi0 --name="${2}" --size="${3}" >/dev/null || perr "cannot create ubifs volume ${2}({$3})"
   else
      pout "create ubifs volume ubi0_${1}:${2}(${3}) ..."
      ubimkvol /dev/ubi0 --vol_id="${1}" --name="${2}" --size="${3}" >/dev/null || perr "cannot create ubifs volume ubi0_${1}:${2}({$3})"
   fi
}

ubi_rmvol () {
   if ubi_exists "${1}"; then
      pout "remove ubifs volume ubi0:${1} ..."
      ubirmvol /dev/ubi0 --name "${1}" || "cannot remove ubifs volume ubi0:${1}"
   fi
}

ubi_detach () {
   pout "detach block device ${MTD_UBI} as ubi device /dev/ubi0 ..."
   ubidetach --dev-path="${MTD_UBI}" || perr "detach block device ${MTD_UBI} as ubi device /dev/ubi0"
}

ubi_mount () {
   pout "mount ubi0:${1} to ${2} ..."
   mount -t ubifs "ubi0:${1}" "${2}" || perr "cannot mount ubi0:${1} to ${2}"
}

ubi_umount () {
   pout "umount ${1} ..."
   umount "${1}" || perr "cannot umount ${1}"
}

ubi_install () {
   pout "install ${1} to ${FILE_MNT}${1} ..."
   install -m 0644 "${1}" "${FILE_MNT}${1}" || perr "cannot install ${1} to ${FILE_MNT}${1}"
}

ubi_install_src () {
   pout "install ${1} to ${FILE_MNT}${2} ..."
   install -m 0644 "${1}" "${FILE_MNT}${2}" || perr "cannot install ${1} to ${FILE_MNT}${2}"
}

ubi_install_mode () {
   pout "install ${2} to ${FILE_MNT}${2}:${1} ..."
   install -m ${1} "${2}" "${FILE_MNT}${2}" || pout "cannot install ${2} to ${FILE_MNT}${2}:${1}"
}

ubi_link () {
   local linkname="${1}"
   local linktarget="$(readlink $linkname)"
   if [ "" = "${linktarget}" ]; then
      pout "no linktarget for ${1}, not creating a link!"
      return 0
   fi
   pout "creating symlink ${FILE_MNT}${linkname} to ${linktarget} ..."
   if [ -n "${FILE_MNT}${linktarget}" ]; then
      ln -sf "${linktarget}" "${FILE_MNT}${linkname}" || \
      pout "cannot link ${FILE_MNT}${linkname} to ${linktarget}"
   else
      pout "not creating link, target does not exist ${FILE_MNT}${linktarget}"
   fi
}

# used only during mode=update!
ubi_install_update_root () {
   pout "install ${ARCH}:${FILE_ROOTFS} content to ${FILE_MNT} ..."
   ( gunzip -c "${ARCH}" | tar -xOf - "${FILE_ROOTFS}" | gunzip -c - | tar -C "${FILE_MNT}" -xf - ) || perr "cannot install ${ARCH}:${FILE_ROOTFS} to ${FILE_MNT}"
}

# used only during mode=flash (install)
ubi_install_oem () {
   pout "install ${FILE_OEM} content to ${FILE_MNT} ..."
   ( gunzip -c "${FILE_OEM}" | tar -C "${FILE_MNT}" -xf - ) || perr "cannot install ${FILE_OEM} to ${FILE_MNT}"
}

# used only during mode=flash (install)
ubi_install_root () {
   pout "install ${FILE_ROOTFS} content to ${FILE_MNT} ..."
   ( gunzip -c "${FILE_ROOTFS}" | tar -C "${FILE_MNT}" -xpf - ) || perr "cannot install ${FILE_ROOTFS} to ${FILE_MNT}"
}

# only used during update (not during install)!
ubi_install_config () {
   gunzip -c "${ARCH}" | tar -xOf - "${1}" |
   while read; do
      REPLY="$(echo "${REPLY}" | egrep -o '^([0-9]+)[ ]+([^ ]+)')"
      local mode="$(echo "${REPLY}" | cut -d' ' -f1)"
      local file="$(echo "${REPLY}" | cut -d' ' -f2)"
      if [ -n "${mode}" -a -n "${file}" ]; then
         ubi_install_mode "${mode}" "${file}"
      fi
   done
}

# only used during update (not during install)!
ubi_create_link () {
   gunzip -c "${ARCH}" | tar -xOf - "${1}" |
   while read; do
      REPLY="$(echo "${REPLY}" | egrep -o '^([^#].*)')"
      local linkname="${REPLY}"
      if [ -n "${linkname}" ]; then
         ubi_link "${linkname}"
      fi
   done
}

ubi_check_error () {
   pout "checking for ubi errors ..."
   if dmesg | egrep -q 'UBI error:'; then
      perr "ubi errors detected:\n$(dmesg | egrep 'UBI error:')"
   fi
}

ftp_load () {
   pout "load ${FTP}${1} ..."
   wget -q -P "${BASE}" "${FTP}${1}" || perr "cannot load "${FTP}${1}" to ${BASE}"
}

ftp_load_all () {
   rm -rf "${BASE}"
   mkdir -p "${BASE}"

   ftp_load 'md5sum'
   (
      grep "$(basename "${FILE_SPL}")"    "${BASE}md5sum"
      grep "$(basename "${FILE_UBOOT}")"  "${BASE}md5sum"
      grep "$(basename "${FILE_ENV}")"    "${BASE}md5sum"
      grep "$(basename "${FILE_KNL}")"    "${BASE}md5sum"
      grep "$(basename "${FILE_DTB}")"    "${BASE}md5sum"
      grep "$(basename "${FILE_ROOTFS}")" "${BASE}md5sum"
      grep "$(basename "${FILE_OEM}")"    "${BASE}md5sum"
   ) > "${BASE}md5sum.f"

   for f in $(cut -d' ' -f3 "${BASE}md5sum.f"); do
      ftp_load "${f}"
   done

   pout "check md5sum(s) ..."
   ( cd "${BASE}" && md5sum -sc 'md5sum.f' ) || perr "invalid md5 checksum(s)"
}

get_system () {
   local oob="$(dmesg | grep 'NAND device:' | egrep -o 'OOB size: [0-9]+' | egrep -o '[0-9]+')"

   if [ "${oob}" -eq 64 ]; then
      echo 'bch8'
   elif [ "${oob}" -eq 128 ]; then
      echo 'bch16'
   else
      echo 'none'
   fi
}

env_get_val () {
   fw_printenv -n "${1}" 2>/dev/null
}

env_set_values () {
   pout "update uboot environment with:\n${1}\n..."
   ( echo "${1}" | fw_setenv -s - ) || perr "cannot update uboot environment with:\n${1}"
}

process_config () {
   pout "process gxserve config files ..."
   chroot "${FILE_MNT}" "/usr/bin/coreconf" "--workingroot=/" || perr "cannot process gxserve configuration files"
}

check_warning () {
   ver="$(grep -E 'version=' /etc/version | grep -Eo '[0-9]\.[0-9][0-9]')"
   if awk "BEGIN { exit !(${ver} >= 1.7 && ${ver} < 1.8) }"; then
      echo ":: 9999 Warning: If you configured UPS Custom Thresholds, please reconfigure them now." >> "${FILE_MNT}/var/log/update.log"
   fi
}

#----- validate param --------------------------------------------------

#----- main ------------------------------------------------------------

if [ "${MODE}" = 'flash' ]; then

   #----- nand flash ---------------------------------------------------

   leds_work

   pout ">> Check for compatible HW"
   if [ "$(get_system)" != "${platform_feature_typ}" ]; then
      pout "@@ Invalid platform type '${platform_feature_typ}' for HW type '$(get_system)'"
      sleep 2
      reboot
      exit 1
   fi

   pout "\$\$ ${platform_serial}-${platform_macaddr}-${platform_oem}-${platform_type}"

   pout ">> Load ftp files ..."
   ftp_load_all

   pout ">> Install U-Boot ..."
   for dev in ${MTD_SPL}; do
      dev_write "${dev}" "${FILE_SPL}"
   done
   dev_write "${MTD_UBOOT}" "${FILE_UBOOT}"

   pout ">> Setup uboot environment ..."
   dev_write "${MTD_U_CNF}" "${FILE_ENV}"
   env_set_values "${ENV_MOD_DUAL}"

   pout ">> Install kernel, dtb for string 1, 2 ..."
   dev_write "${MTD_1_KNL}" "${FILE_KNL}"
   dev_write "${MTD_1_DTB}" "${FILE_DTB}"
   dev_write "${MTD_2_KNL}" "${FILE_KNL}"
   dev_write "${MTD_2_DTB}" "${FILE_DTB}"
   dev_erase "${MTD_OOPS}"

   pout ">> Create ubi0 device and volumes ..."
   dev_erase "${MTD_UBI}"
   ubi_format
   ubi_attach

   ubi_mkvol 0 'rootfs' '150MiB'
   ubi_mount 'rootfs' "${FILE_MNT}"
   ubi_install_root
   ubi_install_oem
   ubi_install "${FILE_PLAT}"
   ubi_umount "${FILE_MNT}"

   ubi_mkvol 1 'rootfs2' '150MiB'
   ubi_mount 'rootfs2' "${FILE_MNT}"
   ubi_install_root
   ubi_install_oem
   ubi_install "${FILE_PLAT}"
   ubi_umount "${FILE_MNT}"

   ubi_detach
   ubi_check_error

   # ok
   pok "nand successfully flashed"

   #----- nand flash end -----------------------------------------------

elif [ "${MODE}" = 'update' ]; then

   #----- update -------------------------------------------------------

   leds_work

   [ -z "${ARCH}" ] && perr "cannot find update archive file"

   oem="$(echo "${ARCH}" | sed -e 's/[^0-9]\+//g')"
   if [ "${oem}" != "${platform_oem}" ]; then
      pout "change oem from ${platform_oem} to ${oem}"
      platform_oem="${oem}"
      FILE_OEM="oem-${platform_oem}.tar.gz"
   fi
   unset oem

   cat >/tmp/platform <<EOF
platform_type='${platform_type}'
platform_oem='${platform_oem}'

platform_serial=${platform_serial}
platform_macaddr=${platform_macaddr}

platform_feature_typ='${platform_feature_typ}'
platform_feature_dtb='${platform_feature_dtb}'
platform_feature_ext_dip_switch='${platform_feature_ext_dip_switch}'
EOF

   ROOTFS="$(mount | cut -d' ' -f1,3 | egrep '^ubi0:rootfs[0-9]* /$')"
   ROOTFS="$(echo "${ROOTFS}" | egrep -o 'rootfs[0-9]*')"
   if [ "${ROOTFS}" = "rootfs" ]; then
      STRING=2
      ROOTFS='rootfs2'
   elif [ "${ROOTFS}" = "rootfs2" ]; then
      STRING=1
      ROOTFS='rootfs'
   else
      perr "fatal: cannot detect root filesystem name '${rootfs}'"
   fi

   RESET="$(echo "${REQUEST_URI}" | egrep -o '[\?&](reset)=([^&]+)' | cut -d= -f2 | head -n 1)"
   RESET_NET="$(echo "${REQUEST_URI}" | egrep -o '[\?&](reset_net)=([^&]+)' | cut -d= -f2 | head -n 1)"

   if [ "${STRING}" -eq 1 ]; then
      pout ">> Update string 1"
      dev_update_write "${MTD_1_KNL}" "${FILE_KNL}"
      dev_update_write "${MTD_1_DTB}" "${FILE_DTB}"
   elif [ "${STRING}" -eq 2 ]; then
      pout ">> Update string 2"
      dev_update_write "${MTD_2_KNL}" "${FILE_KNL}"
      dev_update_write "${MTD_2_DTB}" "${FILE_DTB}"
   fi

   pout ">> Recreate ubifs volume ${ROOTFS}"
   ubi_rmvol 'rootfs_new' # legacy
   ubi_rmvol "${ROOTFS}"
   ubi_mkvol 'auto' "${ROOTFS}" '150MiB'

   ubi_mount "${ROOTFS}" "${FILE_MNT}"
   ubi_install_update_root
   ubi_install_src /tmp/platform "${FILE_PLAT}"
   if [ "${RESET}" = 'true' ]; then
      if [ "${RESET_NET}" != "true" ]; then
         ubi_install_config 'reset-configfiles'
      fi
   else
      ubi_install_config 'keep-configfiles'
      ubi_create_link 'keep-linknames'
   fi
   process_config
   ubi_umount "${FILE_MNT}"

   pout ">> Update uboot environment"
   if [ "$(env_get_val 'envver')" != "${ENV_MOD_VER}" ]; then
      env_set_values "${ENV_MOD_DUAL}"
   fi
   if [ "${STRING}" -eq 1 ]; then
      env_set_values 'nand_pre run nand_pre1'
   elif [ "${STRING}" -eq 2 ]; then
      env_set_values 'nand_pre run nand_pre2'
   fi

   pok "nand successfully updated"

   if [ -f "/var/log/update.log" ]; then
      ubi_mount "${ROOTFS}" "${FILE_MNT}"
      ubi_install_mode "664" "/var/log/update.log"
      chown root.www-data "${FILE_MNT}/var/log/update.log"
      check_warning
      ubi_umount "${FILE_MNT}"
   fi

   reboot

   #----- update end ---------------------------------------------------

else
   MODE='update' # for perr functions
   perr "fatal: script must be executed as flashnand.sh or install.sh"
fi
